/*
SQL Performance Test Application
Copyright (C) 2005-2014, James Gibbons
Released under the GNU Lesser General Public License version 2.1
See http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html
*/

#include "qt_sql_app.h"
#include "ui_qt_sql_app.h"

Qt_SQL_App::Qt_SQL_App(QWidget *parent)
	: QMainWindow(parent),
		ui(new Ui::Qt_SQL_AppClass),dbOpen(false),currentMode(0)
{
	ui->setupUi(this);

	QFileInfo fInfo(qApp->applicationFilePath());
	appFilePath=fInfo.absolutePath()+"/";
	appBaseName=fInfo.completeBaseName();

	insertBlob.clear();

	for (int i=0;i<NB_SIZE;i++)
	{
		insertBlob.push_back(char('A'+(i%20)));
	}

	ui->table1View->setModel(&model1);
	ui->table2View->setModel(&model2);

	resetData();

	// some test code for codecs

	//	QTextCodec *utf8=QTextCodec::codecForName("utf8");
	//	QTextCodec *UTF_8=QTextCodec::codecForName("UTF-8");
	//	QTextCodec::setCodecForLocale(utf8);
	//	QTextCodec *locale=QTextCodec::codecForLocale();

	connect(&idleTimer,SIGNAL(timeout()),this,SLOT(runQuery()));

	connect(&displayTimer,SIGNAL(timeout()),this,SLOT(runDisplay()));

	displayTimer.start(100);
}

Qt_SQL_App::~Qt_SQL_App()
{
	idleTimer.stop();
	displayTimer.stop();

	resetData();
	closeDB();
	delete ui;
}

void Qt_SQL_App::createTable(const QString tableName)
{
	if (!dbOpen) return;

	QSqlQuery q(QSqlDatabase::database("test"));

	QString qString;

	switch (ui->dbType->currentIndex())
	{
		case 0:	// MySQL format
		case 1:	// ODCB MySQL format
			qString="CREATE TABLE "+tableName+" ( "
					"AutoIdx INTEGER NOT NULL PRIMARY KEY AUTO_INCREMENT,"
					"Number1 INTEGER,"
					"Number2 INTEGER,"
					"Date1 DATETIME,"
					"Date2 DATETIME,"
					"Text1 VARCHAR(64),"
					"Text2 VARCHAR(64),"
					"Memo1 TEXT,"
					"Blob1 LONGBLOB,"
					"Boolean1 BOOL,"
					"Boolean2 BOOL"
					") default char set utf8";
			break;
		case 2:	// ODCB MS SQL 2005 Express format
			qString="CREATE TABLE "+tableName+" ( "
					"AutoIdx INTEGER NOT NULL IDENTITY PRIMARY KEY,"
					"Number1 INTEGER,"
					"Number2 INTEGER,"
					"Date1 DATETIME,"
					"Date2 DATETIME,"
					"Text1 NVARCHAR(64),"
					"Text2 NVARCHAR(64),"
					"Memo1 NVARCHAR(MAX),"
					"Blob1 VARBINARY(MAX),"
					"Boolean1 BIT,"
					"Boolean2 BIT"
					")";
			break;
		case 3:	// SQLite format
		case 4:
			qString="CREATE TABLE "+tableName+" ( "
					"AutoIdx INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT,"
					"Number1 INTEGER,"
					"Number2 INTEGER,"
					"Date1 DATETIME,"
					"Date2 DATETIME,"
					"Text1 VARCHAR(64),"
					"Text2 VARCHAR(64),"
					"Memo1 TEXT,"
					"Blob1 LONGBLOB,"
					"Boolean1 BOOL,"
					"Boolean2 BOOL"
					")";
			break;
		case 5:	// PostgreSQL
			qString="CREATE TABLE "+tableName+" ( "
					"AutoIdx SERIAL NOT NULL PRIMARY KEY,"
					"Number1 INTEGER,"
					"Number2 INTEGER,"
					"Date1 TIMESTAMP,"
					"Date2 TIMESTAMP,"
					"Text1 VARCHAR(64),"
					"Text2 VARCHAR(64),"
					"Memo1 TEXT,"
					"Blob1 BYTEA,"
					"Boolean1 BOOLEAN,"
					"Boolean2 BOOLEAN"
					")";
			break;
		default:
			return;
	}
	if (!q.exec(qString))
	{
		QSqlError sqlE=q.lastError();
		QString eS=sqlE.text();
		ui->textEdit->insertPlainText("Database Error: "+eS+'\n');
	}
}

void Qt_SQL_App::openDB()
{
	resetData();
	closeDB();

	QSqlDatabase db;

	switch (ui->dbType->currentIndex())
	{
		case 0:
			db = QSqlDatabase::addDatabase("QMYSQL","test");
			db.setHostName(ui->dbHost->text());
			db.setDatabaseName("test");
			db.setUserName(ui->dbUser->text());
			db.setPassword(ui->dbPassword->text());
			dbOpen=db.open();
			break;
		case 1:
			db = QSqlDatabase::addDatabase("QODBC","test");
			db.setHostName(ui->dbHost->text());
			db.setDatabaseName("test");
			db.setUserName(ui->dbUser->text());
			db.setPassword(ui->dbPassword->text());
			dbOpen=db.open();
			break;
		case 2:
			db = QSqlDatabase::addDatabase("QODBC","test");
			db.setHostName(ui->dbHost->text());
			db.setDatabaseName("test");
			db.setUserName(ui->dbUser->text());
			db.setPassword(ui->dbPassword->text());
			dbOpen=db.open();
			break;
		case 3:
		default:
			db = QSqlDatabase::addDatabase("QSQLITE","test");
			db.setDatabaseName(":memory:");
			dbOpen=db.open();
			break;
		case 4:
			db = QSqlDatabase::addDatabase("QSQLITE","test");
			db.setDatabaseName(appFilePath+"test.db");
			dbOpen=db.open();
			if (dbOpen)
			{
				QSqlQuery q(QSqlDatabase::database("test"));
				if (q.exec("PRAGMA synchronous;"))
				{
					q.first();
					int sVal=q.value(0).toInt();
					QString s;
					QTextStream ts(&s);
					s.clear();
					ts << "Value of synchronous is: " << sVal << endl;
					ui->textEdit->insertPlainText(s);
				}
				if (!q.exec("PRAGMA synchronous=OFF;"))
				{
					QString s;
					QTextStream ts(&s);
					s.clear();
					ts << "Could not turn synchronous off: " << db.lastError().text() << endl;
					ui->textEdit->insertPlainText(s);
				}
				if (q.exec("PRAGMA synchronous;"))
				{
					q.first();
					int sVal=q.value(0).toInt();
					QString s;
					QTextStream ts(&s);
					s.clear();
					ts << "New value of synchronous is: " << sVal << endl;
					ui->textEdit->insertPlainText(s);
				}
			}
			break;
		case 5:	// PostgreSQL
			// use ODCB
			db = QSqlDatabase::addDatabase("QODBC","test");
			db.setHostName(ui->dbHost->text());
			db.setDatabaseName("test");
			db.setUserName(ui->dbUser->text());
			db.setPassword(ui->dbPassword->text());
			dbOpen=db.open();
			// use native driver (disabled)
			//			db = QSqlDatabase::addDatabase("QPSQL","test");
			//			db.setHostName(ui->dbHost->text());
			//			db.setDatabaseName("test");
			//			db.setUserName(ui->dbUser->text());
			//			db.setPassword(ui->dbPassword->text());
			//			dbOpen=db.open();
			break;
	}

	if (dbOpen)
	{
		QStringList tables=db.tables();

		for (int i=0;i<tables.count();i++)
			tables[i]=tables[i].toUpper();

		if (tables.indexOf("TABLE1")<0)
			createTable("Table1");

		if (tables.indexOf("TABLE2")<0)
			createTable("Table2");
	}
	else
	{
		QString s;
		QTextStream ts(&s);
		s.clear();
		ts << "Database failed to open: " << db.lastError().text() << endl;
		ui->textEdit->insertPlainText(s);
	}

	resetData();
}

void Qt_SQL_App::closeDB()
{
	if (dbOpen)
	{
		{
			QSqlDatabase db = QSqlDatabase::database("test");
			db.close();
		}
		QSqlDatabase::removeDatabase("test");
		dbOpen=false;
	}
	resetData();
}

void Qt_SQL_App::resetData()
{
	currentMode=0;
	runCount=0;
	lastRunCount=0;
	lastAutoInc=0;
	HRT.HighResReset();
	if (dbOpen)
	{
		if (currentQuery.isActive())
			currentQuery.clear();

		//
		// WARNING: clear out the query when we are done with it
		// calling clear method is not enough
		// it can issue commands to SQL driver when destructed after database is closed
		// this was causing heap corruption due to what appears to be a MySQL driver bug
		//

		currentQuery=nullQuery;

		// all open data models must be set to a null model
		// these two statements will cause a crash if disabled
		// just click on the "Table 1" tab and close the application
		// with table data showing

		model1.setQuery(nullQuery);
		model2.setQuery(nullQuery);

		idleTimer.start(0);
	}
	else
	{
		idleTimer.stop();
	}
	setControls();
}

void Qt_SQL_App::setControls()
{
	ui->openPB->setEnabled(!dbOpen);
	ui->closePB->setEnabled(dbOpen);
	ui->dbType->setEnabled(!dbOpen);
	ui->dbHost->setEnabled(!dbOpen);
	ui->dbUser->setEnabled(!dbOpen);
	ui->dbPassword->setEnabled(!dbOpen);
	bool b=dbOpen && (currentMode==0);
	ui->emptyTable1Button->setEnabled(b);
	ui->emptyTable2Button->setEnabled(b);
	ui->dropTable1Button->setEnabled(b);
	ui->dropTable2Button->setEnabled(b);
	ui->operationGB->setEnabled(dbOpen);
}

void Qt_SQL_App::runDisplay()
{
	QString s;
	QTextStream ts(&s);
	//ts.setFieldWidth(10);
	ts.setRealNumberPrecision(1);
	ts.setRealNumberNotation(QTextStream::FixedNotation);

	s.clear();
	ts << runCount;
	ui->operationsCount->setText(s);

	s.clear();
	ts << lastAutoInc;
	ui->autoIndex->setText(s);

	if (dbOpen)
		ui->dbOpenLED->setValue(1);
	else
		ui->dbOpenLED->setValue(0);

	double t=HRT.HighResTime();

	if (t>1.0)
	{
		HRT.HighResReset();

		s.clear();
		ts << (runCount-lastRunCount)/t;
		ui->operationsSec->setText(s);
		lastRunCount=runCount;
	}
}

bool Qt_SQL_App::checkExecError(QSqlQuery &query)
{
	const bool qResult=query.exec();

	if (!qResult)
	{
		QSqlError sqlE=query.lastError();

		const int eType=sqlE.type();
		QString sType;

		switch (eType)
		{
			default:
				sType="("+QString::number(eType)+") Invalid";
				break;
			case -1:
				sType="(-1) Not Set";
				break;
			case QSqlError::NoError:
				sType="No Error";
				break;
			case QSqlError::ConnectionError:
				sType="Connection Error";
				break;
			case QSqlError::StatementError:
				sType="Statement Error";
				break;
			case QSqlError::TransactionError:
				sType="Transaction Error";
				break;
			case QSqlError::UnknownError:
				sType="Unknown Error";
				break;
		}

		ui->textEdit->insertPlainText("DB Error Type: "+sType
																	+", DB Text: ("+QString::number(sqlE.number())+") "+sqlE.databaseText()
																	+", Driver Text: "+sqlE.driverText()
																	+'\n');
	}

	return qResult;
}

void Qt_SQL_App::runQuery()
{
	// this runs when event loop is idle
	// when not doing anything, call Sleep or usleep to avoid heavy CPU use

	if (!dbOpen)
	{
#if defined(Q_OS_WIN32) || defined(Q_OS_WIN64)
		Sleep(10);
#endif
#if defined(Q_OS_LINUX) || defined(Q_OS_MAC)
		usleep(10);
#endif
		return;
	}

	switch (currentMode)
	{
		case 0:
			// NOP
#if defined(Q_OS_WIN32) || defined(Q_OS_WIN64)
			Sleep(10);
#endif
#if defined(Q_OS_LINUX) || defined(Q_OS_MAC)
			usleep(10);
#endif
			break;
		case 1:
			// read database
			if (!currentQuery.next())
			{
				currentQuery.seek(-1);
				currentQuery.next();
			}
			if (currentQuery.isValid())
			{
				lastAutoInc=currentQuery.value(0).toULongLong();
				int n1;
				n1=currentQuery.value(1).toInt();
				int n2;
				n2=currentQuery.value(2).toInt();
				QDateTime dt1=currentQuery.value(3).toDateTime();
				QDateTime dt2=currentQuery.value(4).toDateTime();
				QString dtS=dt1.toString();
				QString s1=currentQuery.value(5).toString();
				QString s2=currentQuery.value(6).toString();
				QString s3=currentQuery.value(7).toString();
				QByteArray blob=currentQuery.value(8).toByteArray();
				bool b1;
				b1=currentQuery.value(9).toBool();
				bool b2;
				b2=currentQuery.value(10).toBool();
				runCount++;
			}
			else
			{
				ui->textEdit->insertPlainText("No records to read!\n");
				ui->noneButton->setChecked(true);
			}
			break;
		case 2:
			// insert records
			{
				int n1=rand();
				int n2=rand();
				QDateTime dt1=QDateTime::currentDateTime();
				QDateTime dt2=QDateTime::currentDateTime();

				// some unicode testing

				wchar_t wc[]=L", Russian: \x0441\x0432\x0435\x0434\x0435\x043d\x0438\x0439";
				QString rCode=QString::fromWCharArray(wc);

				QString s1=QDateTime::currentDateTime().toString()+rCode;
				QString s2=QDateTime::currentDateTime().toString()+rCode;
				QString s3=QDateTime::currentDateTime().toString()+rCode;

				int b1=rand() & 1;
				int b2=rand() & 1;

				QSqlQuery insertQuery(QSqlDatabase::database("test"));
				insertQuery.prepare("INSERT INTO Table1 "
														"(Number1,Number2,Date1,Date2,Text1,Text2,Memo1,Blob1,Boolean1,Boolean2) "
														"VALUES "
														"(:n1,:n2,:dt1,:dt2,:s1,:s2,:sl,:blob,:b1,:b2)");
				insertQuery.bindValue(":n1",n1);
				insertQuery.bindValue(":n2",n2);
				insertQuery.bindValue(":dt1",dt1);
				insertQuery.bindValue(":dt2",dt2);
				insertQuery.bindValue(":s1",s1);
				insertQuery.bindValue(":s2",s2);
				insertQuery.bindValue(":sl",s3);
				insertQuery.bindValue(":blob",insertBlob);
				insertQuery.bindValue(":b1",b1);
				insertQuery.bindValue(":b2",b2);

				checkExecError(insertQuery);

				if (lastAutoInc==0)
				{
					insertQuery.exec("SELECT MAX(AutoIdx) FROM Table1");
					insertQuery.first();
					lastAutoInc=insertQuery.value(0).toULongLong();
				}
				else
					lastAutoInc++;

				runCount++;
			}
			break;
		case 3:
			// copy records
			if (!currentQuery.next())
			{
				// when query is done, assume Table1 is done and delete it
				QSqlQuery deleteQuery(QSqlDatabase::database("test"));
				deleteQuery.prepare("DELETE FROM Table1");
				checkExecError(deleteQuery);
				ui->noneButton->setChecked(true);
			}
			if (currentQuery.isValid())
			{
				lastAutoInc=currentQuery.value(0).toULongLong();
				int n1=currentQuery.value(1).toInt();
				int n2=currentQuery.value(2).toInt();
				QDateTime dt1=currentQuery.value(3).toDateTime();
				QDateTime dt2=currentQuery.value(4).toDateTime();
				QString s1=currentQuery.value(5).toString();
				QString s2=currentQuery.value(6).toString();
				QString s3=currentQuery.value(7).toString();
				QByteArray blob=currentQuery.value(8).toByteArray();
				int b1=currentQuery.value(9).toInt();
				int b2=currentQuery.value(10).toInt();

				QSqlQuery insertQuery(QSqlDatabase::database("test"));
				insertQuery.prepare("INSERT INTO Table2 "
														"(Number1,Number2,Date1,Date2,Text1,Text2,Memo1,Blob1,Boolean1,Boolean2) "
														"VALUES "
														"(:n1,:n2,:dt1,:dt2,:s1,:s2,:sl,:blob,:b1,:b2)");
				insertQuery.bindValue(":n1",n1);
				insertQuery.bindValue(":n2",n2);
				insertQuery.bindValue(":dt1",dt1);
				insertQuery.bindValue(":dt2",dt2);
				insertQuery.bindValue(":s1",s1);
				insertQuery.bindValue(":s2",s2);
				insertQuery.bindValue(":sl",s3);
				insertQuery.bindValue(":blob",blob);
				insertQuery.bindValue(":b1",b1);
				insertQuery.bindValue(":b2",b2);
				checkExecError(insertQuery);

				runCount++;
			}
			break;
	}
}

void Qt_SQL_App::on_tabWidget_currentChanged(int tabN)
{
	switch (tabN)
	{
		case 0:
			if (!dbOpen) break;
			//
			// WARNING: calling model.clear() is not good enough and will
			// cause exceptions when the database is closed and opened
			// instead install a null query when not active
			model1.setQuery(nullQuery);
			model2.setQuery(nullQuery);
			break;
		case 1:
			if (!dbOpen) break;
			model1.setQuery("SELECT * FROM Table1",QSqlDatabase::database("test"));
			ui->table1View->show();
			break;
		case 2:
			if (!dbOpen) break;
			model2.setQuery("SELECT * FROM Table2",QSqlDatabase::database("test"));
			ui->table2View->show();
			break;
	}
}

void Qt_SQL_App::on_noneButton_toggled(bool checked)
{
	if (checked)
	{
		resetData();
	}
}

void Qt_SQL_App::on_sequentialReadButton_toggled(bool checked)
{
	if (checked)
	{
		resetData();
		if (dbOpen)
		{
			QSqlQuery q(QSqlDatabase::database("test"));
			currentQuery=q;
			currentQuery.prepare("SELECT * FROM Table1");
			checkExecError(currentQuery);
			currentMode=1;
		}
	}
}

void Qt_SQL_App::on_insertRecordsButton_toggled(bool checked)
{
	if (checked)
	{
		resetData();
		if (dbOpen)
		{
			QSqlQuery q(QSqlDatabase::database("test"));
			currentQuery=q;
			currentMode=2;
		}
	}
}

void Qt_SQL_App::on_copyRecordsButton_toggled(bool checked)
{
	if (checked)
	{
		resetData();
		if (dbOpen)
		{
			QSqlQuery q(QSqlDatabase::database("test"));
			currentQuery=q;
			currentQuery.prepare("SELECT * FROM Table1");
			checkExecError(currentQuery);
			currentMode=3;
		}
	}
}

void Qt_SQL_App::on_emptyTable1Button_clicked()
{
	QSqlQuery q(QSqlDatabase::database("test"));
	q.prepare("DELETE FROM Table1");
	checkExecError(q);
}

void Qt_SQL_App::on_emptyTable2Button_clicked()
{
	QSqlQuery q(QSqlDatabase::database("test"));
	q.prepare("DELETE FROM Table2");
	checkExecError(q);
}

void Qt_SQL_App::on_dropTable1Button_clicked()
{
	{
		QSqlQuery q(QSqlDatabase::database("test"));
		q.prepare("DROP TABLE Table1");
		checkExecError(q);
	}
	createTable("Table1");
}

void Qt_SQL_App::on_dropTable2Button_clicked()
{
	{
		QSqlQuery q(QSqlDatabase::database("test"));
		q.prepare("DROP TABLE Table2");
		checkExecError(q);
	}
	createTable("Table2");
}

void Qt_SQL_App::on_openPB_clicked()
{
	openDB();
}

void Qt_SQL_App::on_closePB_clicked()
{
	ui->noneButton->setChecked(true);
	closeDB();
}
