/*
Qt Threading Test Application
Copyright (C) 2005-2014, James Gibbons
Released under the GNU Lesser General Public License version 2.1
See http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html
*/
#include "Qt_ThreadTest.h"
#include "ui_Qt_ThreadTest.h"
#include <QDebug>

ThreadWorkObject::ThreadWorkObject(QWidget *parent, ThreadObject *setThread,
																	 const int setID)
	:QObject(parent),ourID(setID),ourThread(setThread),runCount(0)
{
}

ThreadWorkObject::~ThreadWorkObject()
{
	runCount=0;
}

void ThreadWorkObject::countValue(void)
{
	QThread *recThread=QThread::currentThread();

	runCount++;

	QString objectID,recvID;

	if (ourThread!=NULL)
	{
		if (recThread==ourThread->guiThread)
			recvID="GUI Recv,    ";
		else if (recThread==ourThread->runThread)
			recvID="Thread Recv, ";
		else
			recvID=QString::number(quint64(recThread))+" Recv, ";

		QMutexLocker locker(&ourThread->lockLists);

		switch (ourID)
		{
			default:
				objectID="X, Unknown";
				break;
			case 0:
				objectID="0, GUI Timer,    GUI Object,    ";
				break;
			case 1:
				objectID="1, GUI Timer,    Thread Object, ";
				break;
			case 2:
				objectID="2, GUI Timer,    Run Object,    ";
				break;
			case 3:
				objectID="3, Thread Timer, GUI Object,    ";
				break;
			case 4:
				objectID="4, Thread Timer, Thread Object, ";
				break;
			case 5:
				objectID="5, Thread Timer, Run Object,    ";
				break;
			case 6:
				objectID="6, Run Timer,    GUI Object,    ";
				break;
			case 7:
				objectID="7, Run Timer,    Thread Object, ";
				break;
			case 8:
				objectID="8, Run Timer,    Run Object,    ";
				break;
		}

		if ((ourID>=0) && (ourID<ourThread->messageList.count()))
		{
			ourThread->messageList[ourID]=objectID+recvID+QString::number(runCount);
		}
	}
}

ThreadObject::ThreadObject(QWidget *parent):
	QThread(parent),
	guiThread(QThread::currentThread()),
	runThread(this),
	gT_gO(new ThreadWorkObject(0,this,0)),
	gT_tO(new ThreadWorkObject(0,this,1)),
	gT_rO(NULL),
	tT_gO(new ThreadWorkObject(0,this,3)),
	tT_tO(new ThreadWorkObject(0,this,4)),
	tT_rO(NULL),
	rT_gO(new ThreadWorkObject(0,this,6)),
	rT_tO(new ThreadWorkObject(0,this,7)),
	rT_rO(NULL),
	guiTimer(new QTimer()),
	threadTimer(new QTimer()),
	runTimer(NULL)
{
	QMutexLocker locker(&lockLists);

	messageList.fill(QString(),9);
	threadList.fill(NULL,9);
	threadList[0]=gT_gO;
	threadList[1]=gT_tO;
	threadList[3]=tT_gO;
	threadList[4]=tT_tO;
	threadList[6]=rT_gO;
	threadList[7]=rT_tO;

	gT_tO->moveToThread(this);
	tT_tO->moveToThread(this);
	rT_tO->moveToThread(this);
	threadTimer->moveToThread(this);

	connect(guiTimer,SIGNAL(timeout()),gT_gO,SLOT(countValue()));
	connect(guiTimer,SIGNAL(timeout()),gT_tO,SLOT(countValue()));
	connect(threadTimer,SIGNAL(timeout()),tT_gO,SLOT(countValue()));
	connect(threadTimer,SIGNAL(timeout()),tT_tO,SLOT(countValue()));

	guiTimer->start(10);
}

ThreadObject::~ThreadObject()
{
	guiTimer->stop();
	delete guiTimer;
	delete threadTimer;

	QMutexLocker locker(&lockLists);

	messageList.clear();
	threadList.clear();

	delete gT_gO;
	delete gT_tO;
	delete tT_gO;
	delete tT_tO;
	delete rT_gO;
	delete rT_tO;
}

void ThreadObject::run()
{
	{
		QMutexLocker locker(&lockLists);

		gT_rO=new ThreadWorkObject(0,qobject_cast<ThreadObject*>(QThread::currentThread()),2);
		tT_rO=new ThreadWorkObject(0,qobject_cast<ThreadObject*>(QThread::currentThread()),5);
		rT_rO=new ThreadWorkObject(0,qobject_cast<ThreadObject*>(QThread::currentThread()),8);

		threadList[2]=gT_rO;
		threadList[5]=tT_rO;
		threadList[8]=rT_rO;
	}

	runTimer=new QTimer();

	connect(guiTimer,SIGNAL(timeout()),gT_rO,SLOT(countValue()));
	connect(threadTimer,SIGNAL(timeout()),tT_rO,SLOT(countValue()));

	connect(runTimer,SIGNAL(timeout()),rT_gO,SLOT(countValue()));
	connect(runTimer,SIGNAL(timeout()),rT_tO,SLOT(countValue()));
	connect(runTimer,SIGNAL(timeout()),rT_rO,SLOT(countValue()));

	threadTimer->start(10);
	runTimer->start(10);

	exec();

	threadTimer->stop();
	runTimer->stop();
	delete runTimer;

	{
		QMutexLocker locker(&lockLists);

		threadList.clear();
	}

	delete gT_rO;
	delete tT_rO;
	delete rT_rO;
}

Qt_ThreadTest::Qt_ThreadTest(QWidget *parent) :
	QMainWindow(parent),
	ui(new Ui::Qt_ThreadTestForm)
{
	ui->setupUi(this);

	connect(&guiDisplayTimer,SIGNAL(timeout()),this,SLOT(runDisplay()));
	guiDisplayTimer.start(100);
	testSignalThread.start();
}

Qt_ThreadTest::~Qt_ThreadTest()
{
	guiDisplayTimer.stop();
	testSignalThread.quit();
	testSignalThread.wait(1000);

	delete ui;
}

void Qt_ThreadTest::runDisplay()
{
	QString displayText;

	{
		QMutexLocker locker(&testSignalThread.lockLists);

		for (int i=0;i<testSignalThread.messageList.count();i++)
		{
			displayText+=testSignalThread.messageList.at(i)+"\n";
		}
	}

	ui->threadWorkMessageList->setPlainText(displayText);
}
